/*
 * rc.c
 *
 *  version 0.10  february 3, 1997
 *  jeremy wohl (jeremyw@anders.com)
 *  no warranty, no responsibility, etc.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/stat.h>
#include <regex.h>

#include "common.h"

#define CACHE_ENTRIES 50


typedef struct {
    bool       service_empty, action_empty, data_empty;
    regex_t    service, action, data;
    char     * command;
} rcline;

typedef struct {
    int       rc_mtime;
    char    * rc_path;
    rcline  * rc_line;
} cacheentry;

static rcline *     rc_getline (FILE *);
static int          split_line (char *, char **, char **, char **, char **);
static void         remove_edgespace (char *);
static void         rc_free (rcline *);
/*static cacheentry * cache_find (const char *);
static void         cache_invalidate (const char *);*/


/* allocates ``*command'' if successful; caller must free
 * returns: 0 if found, 1 if no match, 2 if no rc file, 3 if no such user
 */
int rc_getmatch (const char * user, const char * service, const char * action,
		 const char * data, const char * home, char ** command)
{
    FILE           * fp;
    /*cacheentry     * entry;*/
    struct stat      statbuf;
    rcline         * line;
    char             rcpath [PATH_MAX];
    bool             found = false;

    
    /*
     * check cache
     */
#ifdef 0
    while ((entry = cache_find (user))) {
	if (stat (rcpath, &statbuf) == -1) {
	    cache_invalidate (user);
	    return (2);
	}

	if (statbuf.st_mtime == entry->rc_mtime) {
	    *command = entry->rc_line->command;
	    return (0);
	}
    }
#endif

    /*
     * not cached; has an rc file?
     */
    if (strlen (home) + 1 + 9 + 1 > PATH_MAX)  return (2);

    strcpy (rcpath, home);
    strcat (rcpath, "/");
    strcat (rcpath, ".triggers");

    if (stat (rcpath, &statbuf) == -1)                           return (2);
    if (statbuf.st_mode & S_IWGRP || statbuf.st_mode & S_IWOTH)  return (2);

    /*
     * churn on rc file
     */
    fp = fopen (rcpath, "r");

    while (! found && (line = rc_getline (fp))) {
	if ((line->service_empty || !regexec (&line->service,service,0,0,0)) &&
	    (line->action_empty  || !regexec (&line->action, action, 0,0,0)) &&
	    (line->data_empty    || !regexec (&line->data,   data,   0,0,0)))
        {
	    while ((*command = strdup (line->command)) == 0)  nomem_sleep ();
	    found = true;
	    /* @@ cache_add (user, line); */
	}

	rc_free (line);
    }

    fclose (fp);

    if (! found)
	return (1);

    return (0);
}


/* returns: pointer on success, 0 failure */
rcline * rc_getline (FILE * fp)
{
    rcline  line, * newline;
    char    buf [1024], * service, * action, * data, * command;

    
    while (fgets (buf, 1024, fp)) {
	if (buf [0] == '#')                     /* comment, ignore */
	    continue;

	if (buf [strlen (buf) - 1] != '\n')     /* line too long, ignore */
	    continue;

	remove_edgespace (buf);

	if (buf [0] == 0)                       /* blank line, ignore */
	    continue;

	/* line's ok; try parsing */

	if (split_line (buf, &service, &action, &data, &command) != 0)
	    continue;                           /* badly-formed, ignore */

	remove_edgespace (command);
	
	if (command [0] == 0)                   /* empty command, ignore */
	    continue;

	remove_edgespace (service);
	remove_edgespace (action);
	remove_edgespace (data);

	line.service_empty = service [0] ? false : true;
	line.action_empty  = action  [0] ? false : true;
	line.data_empty    = data    [0] ? false : true;

        #define regc(a,b) (regcomp(a, b, REG_EXTENDED | REG_ICASE | REG_NOSUB))
	if (! line.service_empty && (regc (&line.service, service) != 0))
	    return (0);
	if (! line.service_empty && (regc (&line.action,  action ) != 0)) {
	    regfree (&line.service);
	    return (0);
	}
	if (! line.service_empty && (regc (&line.data,    data   ) != 0)) {
	    regfree (&line.service);
	    regfree (&line.action);
	    return (0);
	}

	while ((line.command = strdup (command))    == 0)  nomem_sleep ();
	while ((newline = malloc (sizeof (rcline))) == 0)  nomem_sleep ();

	memcpy (newline, &line, sizeof (rcline));

	return (newline);
    }

    return (0);
}


void rc_free (rcline * line)
{
    if (! line->service_empty)  regfree (&line->service);
    if (! line->action_empty )  regfree (&line->action);
    if (! line->data_empty   )  regfree (&line->data);

    free (line->command);
    free (line);

    return;
}


/*
 * points pointers to four strings in ``buf'' separated along colons (':').
 * returns: 0 success, 1 bad string
 */
int split_line (char * buf, char ** one, char ** two, char ** three,
		char ** four)
{
    char  * front = buf, * back = buf;
    int     p = 1;


    while (*back) {
	if (*back == ':') {
	    *back = 0;

	    if (p == 1) *one   = front;
	    if (p == 2) *two   = front;
	    if (p == 3) { *three = front; *four = back + 1; return (0); }

	    front = back + 1;
	    p++;
	}

	back++;
    }

    return (1);
}


/* removes leading or trailing space [ \t\n] in-place */
void remove_edgespace (char * string)
{
    char  * hold = string, * front, * back;


    if (*string == 0) return;

    front = hold;
    back  = hold + strlen (string) - 1;

    while (*front == ' ' || *front == '\t' || *front == '\n') front++;

    while (front < back && (*back == ' ' || *back == '\t' || *back == '\n'))
	back--;

    if (back - front == -1)
	*string = 0;
    else {
	memmove (string, front, back - front + 1);
	*(string + (back - front + 1)) = 0;
    }
    
    return;
}
