/*
 * remote.c
 *
 *  version 0.41  april 21, 1998
 *  jeremy wohl (jeremyw@evantide.com)
 *  no warranty, no responsibility, etc.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>

#include "common.h"

typedef struct _alarm_detail_t {
    trigger_t  * trigger;
    int          sockid;
} alarm_detail_t;

pthread_key_t  timeout_alarm_key, deferred_alarm_key;
pthread_key_t  alarm_key;
extern void  * rtriggers;
void         * deferredtriggers;

#define MAX_NS 2000

void * send_trigger (void * arg);
int    read_ns  (int, char *, int);
int    open_socket (char *, int);
void   alarm_handler (int);


void * remote (void * arg)
{
    struct sigaction sa;
    pthread_attr_t   attr;
    pthread_t        thr;
    trigger_t      * trigger;
    

    pthread_attr_init (&attr);
    pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);

    pthread_key_create (&timeout_alarm_key,  0);
    pthread_key_create (&deferred_alarm_key, 0);

    pthread_setspecific (deferred_alarm_key, (void *) 1);

    sa.sa_handler = alarm_handler;
    sa.sa_flags   = SA_RESTART;
    sigemptyset (&sa.sa_mask);
    sigaction (SIGALRM, &sa, 0);

    deferredtriggers = queue_create ();

    /* @@ deal with pendings */

    alarm (5 * 60);

    while (1) {
	queue_sem_wait (rtriggers);
    
	queue_lock (rtriggers);
	trigger = queue_front (rtriggers);
	queue_remove (rtriggers);
	queue_unlock (rtriggers);

	if (pthread_create (&thr, &attr, send_trigger, (void *) trigger)) 
	    queue_add (deferredtriggers, trigger);
    }
}


void * send_trigger (void * arg)
{
    struct timeval  tv;
    alarm_detail_t  alarm_detail;

    trigger_t  * trigger;
    int          sockid;
    int          tv_sec_len, tv_usec_len, delay_len, hops_len, ns_len;
    int          total_len, total_delay;
    char         tv_sec_str [10], tv_usec_str [10], delay_str [10];
    char         hops_str [10], total_str [10], ns [MAX_NS];


    trigger = (trigger_t *) arg;

    alarm_detail.trigger = trigger;
    alarm_detail.sockid  = 0;
    pthread_setspecific (timeout_alarm_key, &alarm_detail);

    alarm (5 * 60);

    if ((sockid = open_socket (trigger->host, 861)) < 1) {
	if (sockid == -1)   /* denied host */
	    free (trigger);
	else                /* temporary failure; defer */
	    queue_add (deferredtriggers, trigger);

	pthread_exit (0);
    }

    alarm_detail.sockid = sockid;

    gettimeofday (&tv, 0);
    total_delay = trigger->delay + (tv.tv_sec - trigger->recv_sec);

    tv_sec_len  = sprintf (tv_sec_str,  "%i", trigger->time_sec);
    tv_usec_len = sprintf (tv_usec_str, "%i", trigger->time_usec);
    delay_len   = sprintf (delay_str,   "%i", total_delay);
    hops_len    = sprintf (hops_str,    "%i", trigger->hops);

    ns_len = sprintf (ns,
		      "%i:%s,%i:%s,%i:%s,%i:%s,%i:%s,%i:%s,%i:%s,%i:%s,"
		      "%i:%s,",
		      1,                         "1",
		      strlen (trigger->user),    trigger->user,
		      strlen (trigger->service), trigger->service,
		      strlen (trigger->action),  trigger->action,
		      strlen (trigger->data),    trigger->data,
		      tv_sec_len,                tv_sec_str,
		      tv_usec_len,               tv_usec_str,
		      delay_len,                 delay_str,
		      hops_len,                  hops_str);
			 
    if (ns_len > MAX_NS)  /* @@ */
	syslog (LOG_ALERT, "big netstring; something went haywire");

    total_len = sprintf (total_str, "%i:", ns_len);
    
    if (write (sockid, total_str, total_len) != total_len)  goto fail;
    if (write (sockid, ns,        ns_len)    != ns_len)     goto fail;
    if (write (sockid, ",",       1)         != 1)          goto fail;

    if (read_ns (sockid, ns, MAX_NS))  goto fail;
    if (strncmp (ns, "K", 1) != 0)     goto fail;

    alarm (0);
    close (sockid);
    free  (trigger);

    pthread_exit (0);

 fail:
    alarm (0);
    close (sockid);
    queue_add (deferredtriggers, trigger);

    pthread_exit (0);
}


void alarm_handler (int whoa)
{
    struct timeval    tv;
    alarm_detail_t  * alarm_detail;
    pthread_attr_t    attr;
    pthread_t         thr;
    trigger_t       * trigger;
    int               i;


    if (pthread_getspecific (deferred_alarm_key)) {
	gettimeofday (&tv, 0);

	queue_lock (deferredtriggers);

	for (i = 0; i < queue_items (deferredtriggers); i++) {
	    trigger = queue_front (deferredtriggers);
	    queue_remove (deferredtriggers);

	    if (tv.tv_sec - trigger->recv_sec > 24 * 60 * 60) {
		free (trigger);
		continue;
	    }

	    pthread_attr_init (&attr);
	    pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
	    
	    if (pthread_create (&thr, &attr, send_trigger, (void *) trigger)) {
		queue_add (deferredtriggers, trigger);
		break;
	    }
	}

	queue_unlock (deferredtriggers);
	
	alarm (300);
    }

    else if ((alarm_detail = pthread_getspecific (timeout_alarm_key))) {
	pthread_setspecific (alarm_key, 0);
	
	if (alarm_detail->sockid)  close (alarm_detail->sockid);
	queue_add (deferredtriggers, alarm_detail->trigger);
	pthread_cancel (pthread_self ());
    }
}


/* returns: -1, denied host, 0 failure, > 0 socket descriptor */
int open_socket (char * remote_ip, int port)
{
    struct timeval       tv;
    struct sockaddr_in   sa;
    struct hostent       hp;

    char  buf [1024];
    int   sockid, hst_errno;


    tv.tv_sec = 120; tv.tv_sec = 0;

    if (gethostbyname_r (remote_ip, &hp, buf, sizeof buf, &hst_errno) == 0) {
	return (-1);
    }

    memset (&sa, 0, sizeof (sa));
    memcpy (&sa.sin_addr, hp.h_addr, hp.h_length);
    sa.sin_family = hp.h_addrtype;
    sa.sin_port   = htons ((unsigned short) port);

    if ((sockid = socket (hp.h_addrtype, SOCK_STREAM, 0)) == -1) {
	return (0);
    }

    if (connect (sockid, (struct sockaddr *) &sa, sizeof (sa)) == -1) {
	close (sockid);
	return (0);
    }

    /*setsockopt (sockid, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof (tv));
    setsockopt (sockid, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof (tv));*/

    return (sockid);
}


/* returns: 0 valid netstring; size <= buf_len, 1 failure */
int read_ns (int sockid, char * buf, int buf_len)
{
    char lenbuf [10];
    int  ns_len, pos = 0;
    

    while (pos < 10 && read (sockid, lenbuf + pos, 1) == 1 &&
	   isdigit (lenbuf [pos]))  pos++;

    if (pos == 0 || pos == 9 || lenbuf [pos] != ':')  return (1);

    lenbuf [pos] = 0;  ns_len = atoi (lenbuf);
    if (ns_len > buf_len - 1)  return (1);

    if (read (sockid, buf, ns_len) != ns_len)  return (1);
    if (read (sockid, lenbuf, 1) != 1)         return (1);
    if (lenbuf [0] != ',')                     return (1);

    buf [ns_len] = 0;

    return (0);
}
