/*
 * triggerd.c
 *
 *  version 0.41  february 3, 1997
 *  jeremy wohl (jeremyw@evantide.com)
 *  no warranty, no responsibility, etc.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include <semaphore.h>
#include <pwd.h>
#include <grp.h>

#include "triggermsg.h"
#include "common.h"

typedef struct {
    char  * user, * home, * command;
    int     uid, gid;
} job_t;

void      * ltriggers, * rtriggers, * jobs;
pthread_t   local_thr, remote_thr, triggers_thr, jobs_thr;

extern void * remote (void * arg);

void * local (void * arg);
void   local_cleanup (void * arg);
void * process_triggers (void * arg);
void * process_jobs (void * arg);
void   main_cleanup ();

int    run (const job_t * job);
bool   strcpy_sz (char *, const char *, int);
int    split_address (char * address, char ** user, char ** host);
char * make_subst_command (char * command, trigger_t * t);
int    find_var (char * string, char ** start, char ** end);


int main ()
{
    struct sigaction sa;

    FILE  * fp;
    int     fd;


    openlog ("triggerd", 0, LOG_DAEMON);

    fp = fopen ("/var/run/triggerd.pid", "w");
    fprintf (fp, "%i\n", getpid ());
    fclose (fp);

    ltriggers  = queue_create ();
    rtriggers  = queue_create ();
    jobs       = queue_create ();

    /* @@ child signals?? */
    sa.sa_handler = main_cleanup;
    sa.sa_flags   = 0;
    sigemptyset (&sa.sa_mask);
    sigaction (SIGTERM, &sa, 0);
    sigaction (SIGINT,  &sa, 0);

    sa.sa_handler = SIG_IGN;
    sigaction (SIGHUP,  &sa, 0);

    fclose (stdin);
    fclose (stdout);
    fclose (stderr);

    fd = open("/dev/null", O_RDWR);
    dup2 (fd, 0);
    dup2 (fd, 1);
    dup2 (fd, 2);

    if (pthread_create (&local_thr,    0, local,            0))  die_quick (0);
    if (pthread_create (&remote_thr,   0, remote,           0))  die_quick (0);
    if (pthread_create (&triggers_thr, 0, process_triggers, 0))  die_quick (0);
    if (pthread_create (&jobs_thr,     0, process_jobs,     0))  die_quick (0);

    pthread_join (local_thr, 0);
    pthread_join (remote_thr, 0);
    pthread_join (triggers_thr, 0);
    pthread_join (jobs_thr, 0);

    queue_free (jobs);
    queue_free (rtriggers);
    queue_free (ltriggers);

    unlink ("/var/run/triggerd.pid");
    closelog ();

    return (0);
}


void main_cleanup ()
{
    pthread_cancel (local_thr);
    pthread_cancel (remote_thr);
    pthread_cancel (triggers_thr);
    pthread_cancel (jobs_thr);
}


void * local (void * arg)
{
    struct trigger_msg  buf;
    struct timeval      tv;

    trigger_t  * t;
    FILE       * fp;
    int          msqid, result;


    msqid = msgget (0, IPC_CREAT | 0666); /* @@ 0622? */

    fp = fopen ("/var/run/triggerd.msqid", "w");
    fprintf (fp, "%i\n", msqid);
    fclose (fp);

    pthread_setcancelstate (PTHREAD_CANCEL_ENABLE, NULL);
    pthread_cleanup_push (local_cleanup, (void *) msqid);

    while (1) {
	result = msgrcv (msqid, (struct msgbuf *) &buf, sizeof buf, 0, 0);

	pthread_testcancel ();

	if (result == -1 && (errno == E2BIG || errno == EINTR))
	    continue;
	else if (result == -1)  /* @@ */
	    break;

	gettimeofday (&tv, 0);

	while ((t = (trigger_t *) malloc (sizeof (trigger_t))) == 0)
	    nomem_sleep ();

	if (buf.version != 1)  goto fail;

	if (! strcpy_sz (t->user,    buf.user,   sizeof buf.user))   goto fail;
	if (! strcpy_sz (t->service,buf.service,sizeof buf.service)) goto fail;
	if (! strcpy_sz (t->action,  buf.action, sizeof buf.action)) goto fail;
	if (! strcpy_sz (t->data,    buf.data,   sizeof buf.data))   goto fail;

	t->time_sec  = buf.time_sec;
	t->time_usec = buf.time_usec;

	t->delay  = buf.delay;
	t->hops   = buf.hops;

	t->host [0] = 0;
	t->recv_sec = tv.tv_sec;

	queue_add (ltriggers, t);
	continue;

    fail:
	free (t);
    }

    pthread_cleanup_pop (0);
    
    return (0);
}


void local_cleanup (void * arg)
{
    msgctl ((int) arg, IPC_RMID, 0);
    unlink ("/var/run/triggerd.msqid");
}


void * process_triggers (void * arg)
{
    struct passwd  * passwd;
    trigger_t      * trigger;
    char           * command, * user, * host;
    job_t          * job;


    /*if (gethostname (hostname, 100)  == -1)  return (2);
    if (h = gethostbyname (hostname) ==  0)  return (2);*/

    while (1) {
	queue_sem_wait (ltriggers);
    
	queue_lock (ltriggers);
	trigger = queue_front (ltriggers);
	queue_remove (ltriggers);
	queue_unlock (ltriggers);

	if (trigger->hops > 20 || (passwd = getpwnam (trigger->user)) == 0 ||
	    passwd->pw_uid == 0)
        {
	    free (trigger);
	    continue;
	}

	if (rc_getmatch (trigger->user, trigger->service, trigger->action,
			 trigger->data, passwd->pw_dir, &command) == 0)
        {
	    if (strncmp ("run ", command, 4) == 0) {
		while (! (job =   malloc (sizeof (job_t))))     nomem_sleep ();
		while (! (job->user = strdup (trigger->user)))  nomem_sleep ();
		while (! (job->home = strdup (passwd->pw_dir))) nomem_sleep ();

		job->command = make_subst_command (command + 4, trigger);
		job->uid = passwd->pw_uid;
		job->gid = passwd->pw_gid;

		queue_add (jobs, job);
	    }

	    else if (strncmp ("passto ",  command, 7) == 0) {
		if (split_address (command + 7, &user, &host))
		    free (trigger);
		else {
		    /* @@ send locally if host is me */

		    if (user [0])  strcpy (trigger->user, user);
		    strcpy (trigger->host, host);
		    
		    queue_add (rtriggers, trigger);
		}
	    }

	    free (command);
	}

	else
	    free (trigger);
    }

    return (0);
}


void * process_jobs (void * arg)
{
    job_t * job;
    int     extant_jobs = 0;


    while (1) {
	if (extant_jobs == 0)
	    queue_sem_wait (jobs);
	else {
	    if (queue_sem_trywait (jobs) == -1) {
		sleep (1);

		while (extant_jobs && waitpid (-1, 0, WNOHANG) > 0)
		    extant_jobs--;

		continue;
	    }
	}

	queue_lock (jobs);
	job = queue_front (jobs);
	queue_remove (jobs);
	queue_unlock (jobs);

	run (job);

	free (job->user);
	free (job->home);
	free (job->command);
	free (job);

	extant_jobs++;
    }

    return (0);
}


int run (const job_t * job)
{
    int pid;


    if ((pid = fork ()) == 0) {
	/* child, ok */

	setenv("USER",  job->user, 1);
	setenv("HOME",  job->home, 1);
	setenv("SHELL", "/bin/sh", 1);

	if (initgroups (job->user, job->gid) == -1)  exit (0);
	if (setregid (job->gid, job->gid)    == -1)  exit (0);
	if (setreuid (job->uid, job->uid)    == -1)  exit (0);

	if (chdir (job->home) == -1)  exit (0);

	execl ("/bin/sh", "/bin/sh", "-c", job->command, 0, 0);

	/* here on failure */
	exit (0);
    }
    if (pid == 0) {
	/* parent, ok */
    }
    if (pid == -1) {
	/* parent, error */
    }

    return (0);
}


char * make_subst_command (char * command, trigger_t * t)
{
    char  * new_command, * old_pos = command, * new_pos, * start, * end;
    int     var_index, len;
    char    tmp [20];
    

    /* @@ size: the maximum allowable config line + current trigger vars */
    while (! (new_command = malloc (sizeof (trigger_t) + 1024)))
	nomem_sleep ();

    new_pos = new_command;
    end     = command;

    while ((var_index = find_var (old_pos, &start, &end))) {
 	memcpy (new_pos, old_pos, start - old_pos);
	new_pos += start - old_pos;
	old_pos  = end;
	
	switch (var_index) {
	case 1:
	    len = strlen (t->user); memcpy (new_pos, t->user, len);
	    new_pos += len; break;
	case 2:
	    len = strlen (t->service); memcpy (new_pos, t->service, len);
	    new_pos += len; break;
	case 3:
	    len = strlen (t->action); memcpy (new_pos, t->action, len);
	    new_pos += len; break;
	case 4:
	    len = strlen (t->data); memcpy (new_pos, t->data, len);
	    new_pos += len; break;
	case 5:
	    len = sprintf (tmp, "%i.%i", t->time_sec, t->time_usec);
	    memcpy (new_pos, tmp, len); new_pos += len; break;
	case 6:
	    len = sprintf (tmp, "%i", t->delay);
	    memcpy (new_pos, tmp, len); new_pos += len; break;
	case 7:
	    len = sprintf (tmp, "%i", t->hops);
	    memcpy (new_pos, tmp, len); new_pos += len; break;
	}
    }

    len = strlen (end);
    memcpy (new_pos, end, len);
    new_pos += len;
    *new_pos = 0;

    return (new_command);
}


/* returns: 0 found one, 1 none found */
int find_var (char * string, char ** start, char ** end)
{
    char  * pos = string, * eos, * start_hold;
    int     var_pos, var_index;
    char    var [20];


    eos = string + strlen (string);

    while ((pos = strchr (pos, '$'))) {
	if (pos > string && *(pos - 1) == '\\') {
	    pos++;
	    continue;
	}

	start_hold = pos;

	pos++;
	var_pos = 0;
	while (pos < eos && var_pos < 20 && isalpha (*pos))
	    var [var_pos++] = *pos++;
	var [var_pos] = 0;

	if      (strcmp (var, "user")      == 0) { var_index = 1; }
	else if (strcmp (var, "service")   == 0) { var_index = 2; }
	else if (strcmp (var, "action")    == 0) { var_index = 3; }
	else if (strcmp (var, "data")      == 0) { var_index = 4; }
	else if (strcmp (var, "timestamp") == 0) { var_index = 5; }
	else if (strcmp (var, "delay")     == 0) { var_index = 6; }
	else if (strcmp (var, "hops")      == 0) { var_index = 7; }
	else
	    continue;

	*start = start_hold;
	*end   = pos;

	return (var_index);
    }

    return (0);
}


/* returns: 0 ok, 1 bad address */
int split_address (char * address, char ** user, char ** host)
{
    char  * result;
    int     len = strlen (address);


    if (*address == 0)  return (1);

    if ((result = strchr (address, '@')) == 0) {
	*user   = address + len;
	*host   = address;
    }
    else {
	*result = 0;
	*user   = address;
	*host   = result + 1;
    }

    return (0);
}


void nomem_sleep (void)
{
    /* @@ is syslog thread-safe? */
    syslog (LOG_ALERT, "cannot alloc memory; sleeping for 5 seconds");
    sleep (5);
}


void die_quick (int kind)
{
    switch (kind) {
    case 0: syslog (LOG_ALERT, "cannot create threads; exiting");  break;
    case 1: syslog (LOG_ALERT, "cannot create sockets; exiting");  break;
    }

    exit (1);
}


bool strcpy_sz (char * dest, const char * src, int max)
{
    max++;

    while (max-- && (*dest++ = *src++) != '\0')
	;

    if (*(dest - 1) != 0)  return (false);  else return (true);
}
